#ifndef CRASH_LEXER
#define CRASH_LEXER

#include <string>
#include <cstring>
#include "ref_ptr.h"

/*
	09/04/02		Created
*/

/**
	Lexer is a class that decomposes strings into their lexical components.

	The Lexer class contains a list of patterns to match against a string,
	which are used by Lexer::iterator to obtain the components.

	If no patterns match, a single character is returned and the iterator::type() 
	will be the character code.

	In addition, the Lexer can transform the input stream as it analyses.

	The general use is:

	Lexer lex;

		for (Lexer::iterator i = lex.begin("this is a string"); i != lex.end(); ++i)
			cout << *i << endl;

	There are two ways to extend a Lexer.
	1.	Subclass and re-implement Lexer::Pattern::match() and Lexer::add().
		Lexer::add() should insert an object of the subclassed Pattern
		into the Lexer with Lexer::addPattern(). All other logic is handled
		by the base Lexer class. This method is useful for generic Lexers
		such as RegexLexer.
	2.	Subclass Lexer::get() and do matching manually. This method should
		set msource to the end of the matched pattern, mmatch to the
		beginning of the matched pattern and mtype to the index of the matched
		pattern. This is useful for hard-coded Lexer's.
*/
class Lexer {
	public :
		class iterator {
			public :
				friend class Lexer;
				iterator() : mlexer(0), msource(0), mmatch(0) {}

				int operator == (iterator const &other) const {
					return msource == other.msource;
				}
				int operator != (iterator const &other) const {
					return msource != other.msource;
				}

				int operator == (char const *str) const { return mstr == str; }
				int operator != (char const *str) const { return mstr != str; }
				int operator < (char const *str) const { return mstr < str; }
				int operator > (char const *str) const { return mstr > str; }
				int operator <= (char const *str) const { return mstr <= str; }
				int operator >= (char const *str) const { return mstr >= str; }

				int operator == (string const &str) const { return mstr == str; }
				int operator != (string const &str) const { return mstr != str; }
				int operator < (string const &str) const { return mstr < str; }
				int operator > (string const &str) const { return mstr > str; }
				int operator <= (string const &str) const { return mstr <= str; }
				int operator >= (string const &str) const { return mstr >= str; }

				int operator [] (int index) const { return mmatch ? mmatch[index] : -1; }

				iterator &operator ++ () { get(); return *this; }
				iterator operator ++ (int) { iterator i = *this; get(); return i; }

				///	The current line number
				int line() const { return mline; }
				///	The length of the SOURCE segment
				int size() const { return msize; }
				///	The indexed type of the segment
				int type() const { return mtype; }

				///	Return a pointer to the segment that this iterator matches
				char const *segment() const { return mmatch; }

				///	Get token (possibly NOT the segment if transform occurred)
				string const &operator * () const { return mstr; }
				///	Get token (possibly NOT the segment if transform occurred)
				string const &value() const { return **this; }
			private :
				iterator(Lexer *lexer, char const *source) 
					: mlexer(lexer), msource(source), mmatch(0), mline(1), msize(0), mtype(-1) {
					get();
				}

				inline void get() { mlexer->preget(mstr, msource, mmatch, mtype, msize, mline); }

				Lexer *mlexer;
				char const *msource, *mmatch;
				int mline, msize, mtype;
				mutable string mstr;
		};
		friend class iterator;

		Lexer(int maxpatterns = 256);
		virtual ~Lexer();

		///	Start lexical analysis of the specific string
		iterator begin(char const *str) { return iterator(this, str); }
		static iterator end() { return iterator(0, 0); }

		/**	Add a new pattern to the Lexer. Subclasses should overload this
			to insert their subclassed Pattern's. The indexes of a 
			Lexer should start at 256.
		*/
		virtual void add(int index, char const *pattern, 
			bool ignore = false, bool enable = true);
		void ignore(int index, char const *pattern, bool enable = true) {
			add(index, pattern, true, enable);
		}

		///	Ignore the specific pattern index
		void ignore(int index, bool state = true) { mpattern[index - 256]->ignored = state; }
		///	Enable the specific pattern index
		void enable(int index, bool state = true) { mpattern[index - 256]->enabled = state; }
		bool ignored(int index) const { return mpattern[index]->ignored; }
		bool enabled(int index) const { return mpattern[index]->enabled; }

	protected :
		/**	A Pattern is the basic matching component of a Lexer. It can be
			subclassed by Lexer subclasses that do not wish to overload
			Lexer::get().
		*/
		struct Pattern {
			Pattern(bool ignored = false, bool enabled = false) : ignored(ignored), enabled(enabled) {}
			virtual ~Pattern() {}

			///	Returns the size of the match against the parameter, or -1
			virtual int match(char const *) { return -1; }

			bool ignored, enabled;
		};

		enum MatchResult {
			MATCH_NONE,
			MATCH_SOURCE,
			MATCH_MUTATED,
		};

		virtual MatchResult get(string &out, char const *&end, char const *&start, int &mtype);
		void addPattern(int index, Pattern *pattern);

	private :
		bool preget(string &out, char const *&msource, char const *&mmatch, int &mtype, int &msize, int &mline);

		ref_ptr<Pattern> *mpattern;
		int mpatterns;
};

#endif
