#include <cstring>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <climits>
#include <cstdio>
#include <iostream>
#include <string>
#include <vector>
#include "config.h"
#include "String.h"
#include "RCParser.h"
#include "main.h"

using namespace std;

int main(int argc, char **argv) {
bool verbose = false, onenterleave = false;
Shell shell = BASH;
char *olddir = argv[1], *newdir = 0;
RCParser rc;

	try {
		rc("/etc/ondirrc");
		if (getenv("HOME")) {
		string home = string(getenv("HOME")) + "/.ondirrc";

			rc(home);
		}
	} catch (RCParser::exception &e) {
		cerr << "error:" << e.line() << ": " << e.what() << endl;
		return 1;
	}

//	if (getuid() == 0) {
//		cerr << argv[0] << ": error, refusing to run as root" << endl;
//		return 1;
//	}

	// parse options
	{
	int i;
		for (i = 1; i < argc && argv[i][0] == '-'; ++i)
			switch (argv[i][1]) {
				case 'V' :
					cout << VERSION << endl;
					return 0;
				break;
				case 'o' : // onenter/onleave
					onenterleave = true;
				break;
				case 'v' : // verbose
					verbose = true;
				break;
				case 't' : // tcsh
					shell = TCSH;
				break;
				default :
					cerr << argv[0] << ": error, unknown option '" << argv[i] << "'" << endl;
				case 'h' :
					cout << "usage: " << argv[0] << " [-v] [-t] [-h] <old directory> [<new directory>]" << endl
						<< "  -v    verbose" << endl
						<< "  -t    display TCSH commands rather than BASH" << endl
						<< "  -h    help" << endl
						<< "  -V    show version" << endl
						<< "  -o    honour .onenter/.onleave scripts" << endl;
					return 0;
				break;
			}
		if (i >= argc) {
			cerr << argv[0] << ": error, path must be specified" << endl;
			return 1;
		}
		olddir = argv[i++];
		if (i < argc)
			newdir = argv[i];
	}

	// get base working directories
vector<string> last_dir, current_dir;

	{
	char tmp[strlen(olddir) + 1];
		strcpy(tmp, olddir);
		if (tmp[0] != '/') {
			cerr << argv[0] << ": error, path must be absolute" << endl;
			return 1;
		}
		while (tmp[strlen(tmp) - 1] == '/') tmp[strlen(tmp) - 1] = 0;
		last_dir = split(tmp);
	}
	{
	char tmp[PATH_MAX];

		if (newdir)
			strcpy(tmp, newdir);
		else
			getcwd(tmp, PATH_MAX);
		while (tmp[strlen(tmp) - 1] == '/') tmp[strlen(tmp) - 1] = 0;
		current_dir = split(tmp);
	}

unsigned common_root;

	// Find common root path
	for (common_root = 0; last_dir.size() > common_root && current_dir.size() > common_root && last_dir[common_root] == current_dir[common_root] ; common_root++) ;
	if (common_root) common_root--;

	// Move up to common root from old path
	for (unsigned i = last_dir.size() - 1; i > common_root; i--) {
	string leaving = join(last_dir, i);

		if (rc.leavePath(leaving) != "") {
		string code = rc.leavePath(leaving);

			if (verbose) cerr << "ondir: executing leave code for '" << leaving << "'" << endl;
			execute(shell, leaving, code);
		} else
		if (onenterleave) {
		string filename = leaving + "/.onleave";

			if (!readable(filename))
				if (readable(leaving + "/.onexit"))
					filename = leaving + "/.onexit";
			if (readable(filename)) {
				if (!goodpermissions(filename))
					cerr << argv[0] << ": warning, not executing '" << filename << "' as it has bad permissions/ownership" << endl;
				else {
					if (verbose) cerr << argv[0] << ": called exit script in '" << leaving << "'" << endl;
					execute(shell, leaving, filename, true);
				}
			}
		}
	}
	// Move in to new path from common root
	for (unsigned i = common_root + 1; i < current_dir.size(); i++) {
	string entering = join(current_dir, i);

		if (rc.enterPath(entering) != "") {
		string code = rc.enterPath(entering);

			if (verbose) cerr << "ondir: executing enter code for '" << entering << "'" << endl;
			execute(shell, entering, code);
		} else
		if (onenterleave) {
		string filename = entering + "/.onenter";

			if (readable(filename)) {
				if (!goodpermissions(filename))
					cerr << argv[0] << ": warning, not executing '" << filename << "' as it has bad permissions/ownership" << endl;
				else {
					if (verbose) cerr << argv[0] << ": called enter script in '" << entering << "'" << endl;
					execute(shell, entering, filename, true);
				}
			}
		}
	}
	return 0;
}

void execute(Shell shell, string const &dir, string const &code, bool file = false) {
	if (shell == TCSH) {
		cout << "setenv ONDIRWD \"" << dir << "\";" << endl;
		if (file)
			cout << "source \"" << code << "\";" << endl;
		else
			cout << str::replace("\n", ";\n", str::trim(code)) << ";" << endl;
		cout << endl << "unsetenv ONDIRWD;" << endl;
	} else {
		cout << "export ONDIRWD=\"" << dir << "\";" << endl;
		if (file)
			cout << ". \"" << code << "\";" << endl;
		else
			cout << str::trim(code) << endl;
		cout << endl << "unset ONDIRWD;" << endl;
	}
}

/* Split a path into a vector */
vector<string> split(char const *str) {
vector<string> out;
char const *last;

	if (str[0] == '/') str++;
	for (last = str; *str; str++) {
		if (*str == '/') {
			out.push_back(string(last, str - last));
			last = str + 1;
		}
	}
	out.push_back(string(last, str - last));
	return out;
}

/* Join a split path into a string */
string join(vector<string> const &in, unsigned to) {
string out;

	for (unsigned i = 0; i <= to; ++i)
		out += "/" + in[i];
	return out;
}

/* Return true if file is readable */
bool readable(string const &file) {
struct stat s;

	if (!stat(file.c_str(), &s) && 
		s.st_mode & (S_IRUSR | S_IRGRP | S_IROTH))
			return true;
	return false;
}

/* Return true if file has good permissions */
bool goodpermissions(string const &file) {
struct stat s;

	if (!stat(file.c_str(), &s) && s.st_uid == getuid() &&
		!(s.st_mode & (S_IWGRP | S_IWOTH)))
			return true;
	return false;
}

void enterPath(string const &path) {
}

void exitPath(string const &path) {
}

string expandString(string const &str) {
string out;

	for (char const *i = str.c_str(); *i != 0; ++i)
		if (*i == '$') {
		string word;

			++i;
			while (isalnum(*i) || *i == '_') {
				word += *i;
				++i;
			}
			--i;
			if (getenv(word.c_str()))
				out += getenv(word.c_str());
		} else
			out += *i;
	return out;
}
