#ifndef CRASH_REF_PTR
#define CRASH_REF_PTR

/**
	ref_ptr<T> is a class that manages reference counting for a pointer.

	ref_ptr<string> mname;

	This will declare a reference counted pointer to a string. 
	
	Note that the actual type is given without a pointer * after it.
*/
template <typename T>
class ref_ptr {
		class reference {
			public :
				reference(T *t, int refs = 1) : t(t), refs(refs) {}
				~reference() { deref(); }

				T *get() { return t; }

				int references() const { return refs; }

				bool deref() { if (--refs <= 0) { delete t; t = 0; return true; } return false; }
				void ref() { refs++; }
			protected :
				T *t;
				int refs;
		};

	public :
		///	Construct an empty ref_ptr
		ref_ptr() : mref(0) {}
		///	Construct a ref_ptr that manages t
		ref_ptr(T *t, int refs = 1) : mref(new reference(t, refs)) {}
		///	Copy a reference counted pointer
		ref_ptr(ref_ptr const &copy) : mref(0) { *this = copy; }
		///	Dereference and possibly destroy a pointer
		~ref_ptr() { deref(); }

		///	Assign another pointer to this one
		ref_ptr &operator = (ref_ptr const &copy) {
			if (copy.mref == mref) return *this;
			deref(); mref = copy.mref; ref();
			return *this;
		}

		///	Assign a raw pointer to be reference counted
		ref_ptr &operator = (T *t) {
			if (mref) {
				if (t == mref->get()) return *this;
				deref();
			}
			mref = new reference(t);
			return *this;
		}

		///	Return the reference counted pointer
		T *operator -> () { return mref ? mref->get() : 0; }
		///	Return the reference counted pointer
		T *operator -> () const { return mref ? mref->get() : 0; }
		///	Return the reference counted pointer
		operator T * () { return mref ? mref->get() : 0; }
		///	Return the reference counted pointer
		operator T * () const { return mref ? mref->get() : 0; }
		///	Return the reference counted pointer
		T &operator * () { return *(mref ? mref->get() : 0); }
		///	Return the reference counted pointer
		T const &operator * () const { return *(mref ? mref->get() : 0); }

		///	Get the number of references to the pointer being held
		int references() const { return mref ? mref->references() : 0; }

		///	Release reference
		void release() { deref(); }

	protected :
		bool deref() { if (mref && mref->deref()) { delete mref; mref = 0; return true; } mref = 0; return false; }
		void ref() { if (mref) mref->ref(); }

	private :
		reference *mref;
};

#endif
